"""Heuristics that allows to find new potential auxiliary points in a problem."""

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup

__all__ = [
    "HeuristicName",
    "Heuristic",
    "HeuristicSetup",
]
